#include <SPI_MSTransfer.h>

SPI_MSTransfer slave = SPI_MSTransfer("SLAVE", "STANDALONE");

#define LINE(name,val) Serial.print(name); Serial.print("\t"); Serial.println(val);

#define data_elements 36

void setup() {
	Serial.begin(115200);
	while (!Serial && millis() < 5000 )
	{	Serial.print( "Teensy NOT Online @ millis=" );
		Serial.println( millis() );
		delay(30);
	}
	Serial.print( "Teensy Online @ millis=" );
	Serial.println( millis() );
	Serial.println("\n" __FILE__ " " __DATE__ " " __TIME__);

	if ( ARM_DWT_CYCCNT == ARM_DWT_CYCCNT ) {
		Serial.print( "Cycle Counter Not Enabled :" );
		Serial.println( ARM_DWT_CYCCNT );
	}
	if ( ARM_DWT_CYCCNT == ARM_DWT_CYCCNT ) {
		// Enable CPU Cycle Count
		ARM_DEMCR |= ARM_DEMCR_TRCENA;
		ARM_DWT_CTRL |= ARM_DWT_CTRL_CYCCNTENA;
	}
	if ( ARM_DWT_CYCCNT != ARM_DWT_CYCCNT ) {
		Serial.print( "Cycle Counter Enabled! :" );
		Serial.println( ARM_DWT_CYCCNT );
	}

  Serial.println("\n" __FILE__ " " __DATE__ " " __TIME__);
  LINE("F_CPU", F_CPU);
  
	slave.begin( );
	slave.onTransfer(myCallback);
	// slave.debug(Serial); // SPI_MST Debug error tracking

}
void led() {
	digitalWrite(13, !digitalRead(13));
}

void loop() {
	slave.events();
}

static uint16_t last_packetID = 0;
elapsedMillis TogClk;
uint32_t TogCnt = 0;
uint32_t FaFCnt = 0;
void myCallback(uint16_t *buffer, uint16_t length, AsyncMST info) {
	if ( 55 == info.packetID ) {
		if ( 4*data_elements != length ) {
			Serial.print("Bad Length: "); Serial.println(length); // If this shows then the SPI Passed array size is nor wrong
		}
		else if ( 0 != info.error ) {
			Serial.println("\nBad CRC: ");
		}
		else {
			double* MST_PrintVals;
			MST_PrintVals = (double *)buffer;

			{
				//trig conversions
				float rad2deg = 180.0f / PI;
				float deg2rad = PI / 180.0f;
				int  textLength = data_elements * 31;
				char text[textLength];

        char rtcTime_text[30];
        //char currentTOW_text[30];
        char pitch_text[30]; char roll_text[30];
        char yaw_text[30]; char heading_text[30];
        char lat_text[30]; char lon_text[30]; char alt_text[30];
        char velN_text[30]; char velE_text[30]; char velD_text[30];
        char gbx_text[30]; char gby_text[30]; char gbz_text[30];
        char abx_text[30]; char aby_text[30]; char abz_text[30];
  
        //Temporay buffers for matrices
        char p_text[30], q_text[30], mat_buf[19*31];

				int ii = 0;
				dtostrf( MST_PrintVals[ii] , 10, 4, rtcTime_text);
        ii++;
        dtostrf( MST_PrintVals[ii], 10, 4, pitch_text);
				ii++;
				dtostrf( MST_PrintVals[ii], 10, 4, roll_text);
				ii++;
				dtostrf( MST_PrintVals[ii], 10, 4, yaw_text);
				ii++;
				dtostrf( MST_PrintVals[ii], 10, 4, heading_text);
        ii++;
        dtostrf( MST_PrintVals[ii] , 10, 4, lat_text);
				ii++;
				dtostrf( MST_PrintVals[ii] , 10, 4, lon_text);
				ii++;
				dtostrf( MST_PrintVals[ii] , 10, 4, alt_text);
				ii++;
				dtostrf(MST_PrintVals[ii], 10, 4, velN_text);
				ii++;
				dtostrf( MST_PrintVals[ii], 10, 4, velE_text);
        ii++;
        dtostrf( MST_PrintVals[ii], 10, 4, velD_text);
        ii++;
        dtostrf( MST_PrintVals[ii] , 10, 10, gbx_text);
        ii++;
        dtostrf( MST_PrintVals[ii] , 10, 10, gby_text);
				ii++;
				dtostrf( MST_PrintVals[ii] , 10, 10, gbz_text);
				ii++;
				dtostrf( MST_PrintVals[ii] , 10, 10, abx_text);
				ii++;
				dtostrf( MST_PrintVals[ii] , 10, 10, aby_text);
        ii++;
        dtostrf( MST_PrintVals[ii] , 10, 10, abz_text);

      mat_buf[0] = '\0';
      //set up sqrt(P) and q matrices for output
      for(int i=0; i < 15; i++){
        ii++;
        dtostrf(MST_PrintVals[ii],10,6, p_text);
        strcat(mat_buf, p_text); strcat(mat_buf, ",");
      }
      for(int i=0; i < 4; i++){
        ii++;
        dtostrf(MST_PrintVals[ii],10,6, q_text);
        strcat(mat_buf, q_text); strcat(mat_buf, ",");
      }

				// Create single text parameter and print it
      snprintf(text, textLength, "%s,%s,%s,%s,%s, %s,%s,%s,%s,%s, %s,%s,%s,%s,%s, %s,%s", 
        rtcTime_text, pitch_text, roll_text, yaw_text,heading_text,
        lat_text, lon_text, alt_text, 
        velN_text, velE_text, velD_text, 
        gbx_text, gby_text, gbz_text, 
        abx_text, aby_text, abz_text );

      Serial.print(text); Serial.println(mat_buf);

				//Serial.send_now();
			}
		}
	}
	else if ( 60 == info.packetID ) {
		if ( 4*data_elements != length ) {
			Serial.print("Bad Length: "); Serial.println(length); // If this shows then the SPI Passed array size is nor wrong
		}
		else if ( 0 != info.error ) {
			Serial.println("\nBad CRC: ");
		}
		else {
			double* MST_PrintVals;
			MST_PrintVals = (double *)buffer;

// This provides a terse output with error checked based on MASTER output to the 12 DOUBLE value Array
			{
				static uint16_t TogLast = 0;
				static uint16_t TogHz =0;
				static uint16_t FaFHz =0;
				static double LastVal = 100000;
				static uint32_t ChkErr = 0;
				static uint32_t CBcount = 0;
				if ( digitalReadFast( LED_BUILTIN) != TogLast ) { 
					TogLast = !TogLast; 
					TogCnt++; 
				}
				if ( TogClk >=1000 ) {
					TogClk -= 1000;
					TogHz = TogCnt;
					TogCnt = 0;
					FaFHz = FaFCnt;
					FaFCnt = 0;
				}
				if ( last_packetID != info.packetID ) {
					LastVal += data_elements;
					if ( LastVal > 65536 && LastVal < 100000 ) LastVal -= 65536;
				}
				elapsedMillis LastCB;
				double DiffMiss = 0;
				uint32_t ErrSeen = 0;
				CBcount++;
				FaFCnt++;
				for ( uint32_t ii = 0; ii < data_elements; ii++ ) {
					LastVal++;
					if ( 65536 == LastVal ) LastVal = 0;
					if ( LastCB > 10 || 100001 == LastVal || 100013 == LastVal ) LastVal = MST_PrintVals[ii];
					if ( LastVal != MST_PrintVals[ii] ) {
						DiffMiss = LastVal - MST_PrintVals[ii];
						ChkErr++;
						Serial.print("\nBad LASTVAL TEST INCREMENT <<<<<<<<<<<<<<<<<<<< DIFF OF> ");
						Serial.println( DiffMiss );
						LastVal = MST_PrintVals[ii];
						ErrSeen++;
					}
					if ( 0 != ErrSeen || 1 > ii ) {
						Serial.print( MST_PrintVals[ii] );
						Serial.print(",");
					}
					else {
						Serial.print(" #,");
					}
				}
				Serial.print( CBcount );
				Serial.print(",");
				Serial.print( ChkErr );
				Serial.print(" [");
				Serial.print( FaFHz );
				Serial.print(" ,");
				Serial.print( TogHz );
				Serial.println();
				LastCB = 0;
			}
		}
	}
	else {
		Serial.print("PacketID: ");
		Serial.println(info.packetID);
	}
	last_packetID = info.packetID;
}
